<?php

namespace Imagine;

class Image extends Canvas {
    const DEFAULT_QUALITY = 90;

    protected $filename = "";

    protected $type = "image/jpeg";

    public function __construct(/* resource|string */ $image = null)
    {
        if ( is_string( $image ) ) {
            $this->filename = $image;
            $this->fromFile($image);
        }
        elseif ( is_resource($image) ) {
            $this->image = $image;
        }
        else {
            parent::__construct();
        }
    }

    public function type($set = null) {
        return $set === null ? $this->type : $this->type = $set;
    }

    public function fromFile($filename) {
        #dump($filename);
        switch ( \exif_imagetype($filename) ) {
            case 1 :
                $this->forceType('image/gif')->image(\imagecreatefromgif($filename));
                break;
            case 2 :
                $this->forceType('image/jpeg')->image($this->loadJpeg($filename));
            break;
            case 3 :
                $this->forceType('image/png')->image(\imagecreatefrompng($filename));
            break;
            case 6 :
                $this->forceType('image/bmp')->image(\imagecreatefrombmp($filename));
            break;
            case 18:
                $this->forceType('image/webp')->image(\imagecreatefromwebp($filename));
            break;
        }

        if ( $this->valid() ) {
            $this->width = \imagesx($this->image);
            $this->height = \imagesy($this->image);
        }

        return $this;
    }

    public function fromString($content) {
        $this->image( \imagecreatefromstring($content) );

        $info = \getimagesizefromstring($content);

        $this->width = $info[0];
        $this->height = $info[1];

        $this->forceType($info['mime']);

        return $this;
    }

    public function forceType(string $type) {
        $this->type = $type;

        return $this;
    }

    public function valid() : bool
    {
        return is_resource($this->image ?? null) || $this->image instanceof \GdImage;
    }

    public function loadJpeg(string $filename) /* : false|GdImage */
    {
        try {
            $img  = \imagecreatefromjpeg($filename);
        }
        catch(\Exception $e) {
            return false;
        }
        
        $exif = @\exif_read_data($filename);

        if ( $img && !empty($exif['Orientation']) ) {
            switch($exif['Orientation']){
                case 6:
                case 5:
                    $img = \imagerotate($img, 270, null);
                    break;

                case 3:
                case 4:
                    $img = \imagerotate($img, 180, null);
                    break;

                case 7:
                case 8:
                    $img = \imagerotate($img, 90, null);
                    break;

            /*
                case 4:
                case 5:
                case 7:
                    \imageflip($img, \IMG_FLIP_HORIZONTAL);
                    break;
            */
            }
        }

        return $img;
    }

    public function save(int $quality = null, $filters = null) : bool
    {
        return $this->render($this->filename, $quality, $filters);
    }

    public function render(string $path = null, int $quality = null, $filters = null) # : bool
    {
        $layers = $this->layers();

        if ($path === null) {
            $path = fopen("php://memory", "w+");
        }

        usort($layers, function($e1, $e2) {
            return $e1->order() <=> $e2->order();
        });

        foreach($layers as $item) {
            $this->merge( $item->image() );
        }

        switch( $this->type ) {
            case "image/gif":
                $retval = \imagegif($this->image(), $path);
                break;

            case "image/webp":
                $retval = \imagewebp($this->image(), $path, $quality ?: static::DEFAULT_QUALITY);
            break;

            case "image/png":
                \imagealphablending($this->image, true);
                \imagesavealpha($this->image, true);
                $retval = \imagepng($this->image(), $path, $quality === null ? round( static::DEFAULT_QUALITY / 10, 0) : $this->normalizeQuality($quality), $filters);
            break;

            default:
            case "image/jpeg":
            case "image/jpg":
                $retval = \imagejpeg($this->image(), $path, $quality ?: static::DEFAULT_QUALITY);
            break;

        }

        if ( is_resource($path) ) {
            rewind($path);
            $retval = stream_get_contents($path);
        }

        return $retval ?? null;
    }

    public function readExif() : array
    {
        return \exif_read_data($this->filename, 0, true);
    }

    public function remove_exif($old, $new) {
        $f1 = fopen($old, 'rb');
        $f2 = fopen($new, 'wb');

        // Find EXIF marker
        while ( $s = fread($f1, 2) ) {
            $word = unpack('ni', $s)['i'];
            if ($word == 0xFFE1) {
                // Read length (includes the word used for the length)
                $s = fread($f1, 2);
                $len = unpack('ni', $s)['i'];
                // Skip the EXIF info
                fread($f1, $len - 2);
                break;
            } else {
                fwrite($f2, $s, 2);
            }
        }

        // Write the rest of the file
        while ( $s = fread($f1, 4096) ) {
            fwrite($f2, $s, strlen($s));
        }

        fclose($f1);
        fclose($f2);
    }

    protected function normalizeQuality($quality) {
        switch($this->type()) {
            case "image/png":
                return round($quality / 100, 0);

            default:
                return $quality;
        }
    }
}
